//
// Copyright (c) 1997,1998 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// See the README.TXT file that came with this software for restrictions
// on the use and redistribution of this file or send E-mail to
// info@colosseumbuilders.com
//

//
//  Title:  PNG Encoder Class Implementation
//
//  Author:  John M. Miano miano@colosseumbuilders.com
//
#include <limits.h>

#include "pngencod.h"
#include "pngpvt.h"
#include "pngchksm.h"
#include "pnghuffe.h"
#include "stdio.h"

// Hash Table size definitions.
const unsigned int HashBits = 5 ;
const unsigned char lenKeyHash = 3; //    -
const unsigned int HashTableSize = 1 << (3 * HashBits) ;
const int maxBlockBufferSize=0x10000;
// PngWindowSize -  PNG- -     
const unsigned int minLenZamina=3; //      <; >
const unsigned int minLenZaminaLiteral=12; //     ,   
extern unsigned long countZamina, countZaminaV, countZaminaG;

//        JPEGLS
#include "klJPEGLS.h"

bool trebaCorectPredict=true;

corectPredict corectComponent[3];

// 
extern generatePaletr(UBYTE1 *imageData, UBYTE4 *nextPoint, UBYTE1 *indexPointPaletr);
extern void initBufPlus();
extern void writeBufPlus(UBYTE1 w, UBYTE1 countBitW);
extern void doneBufPlus();
extern UBYTE2 countColorPaletr;
extern UBYTE1 minRColorPaletr[256], maxRColorPaletr[256], minGColorPaletr[256], maxGColorPaletr[256], minBColorPaletr[256], maxBColorPaletr[256];

extern UBYTE1 countBitRColorPaletr[256], countBitGColorPaletr[256], countBitBColorPaletr[256];

UBYTE4 currentPointWrite;


//        JPEGLS

extern void initializeGlobalMasJPEGLSPredict();

extern void initializeComponentMasJPEGLSPredict();

extern unsigned char JPEGLSPredict(unsigned char a, unsigned char b, unsigned char c, unsigned char d, unsigned char x);


//    
UBYTE4 globalBlockBufferCount, currentGlobalPoz;
UBYTE2 *globalBlockBuffer;
UBYTE1 *indexPointPaletr;

// Size of the output buffer. This value defines the maximum size for
// IDAT block. This value must be larger than the size of any non-IDAT
// blocks used to write the image. Other than that the value can be
// selected arbitrarily.
const unsigned int OutputBufferSize = (1 << 13) ;

static inline Hash (UBYTE1 v1, UBYTE1 v2, UBYTE1 v3)
{
  const unsigned int mask = (1 << HashBits) - 1 ;
  unsigned int value = (v1 & mask )
                     | ((v2 & mask) << HashBits)
                     | ((v3 & mask) << (2 * HashBits)) ;
  return value ;

}

  // extras [n] is the number of extra bits for the code n.
  static const unsigned int extrasDistance [PngMaxDistanceCodes] =
    {
      0, 0,  0,  0,  1,  1,  2,  2,  3, 3,
      4, 4,  5,  5,  6,  6,  7,  7,  8, 8,
      9, 9, 10, 10, 11, 11, 12, 12, 13, 13,
    } ;

static void DistanceToCode (unsigned int distance, unsigned int &code,
                            unsigned int &extra, unsigned int &value)
{
  // maxvalue [n] is the maximum distance value for the code n.
  static const unsigned int maxvalue [PngMaxDistanceCodes] =
    {
         1,     2,     3,     4,     6,     8,    12,    16,    24,    32,
        48,    64,    96,   128,   192,   256,   384,   512,   768,  1024,
      1536,  2048,  3072,  4096,  6144,  8192, 12288, 16384, 24576, 32768,
    } ;

  // bases [n] is the smallest distance value for code n.
  static const unsigned int bases [PngMaxDistanceCodes] =
    {
         1,    2,    3,    4,    5,    7,    9,    13,    17,    25,
        33,   49,   65,   97,  129,  193,  257,   385,   513,   769,
      1025, 1537, 2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577,
    } ;

  for (code = 0 ; code < PngMaxDistanceCodes ; ++ code)
  {
    if (distance <=  maxvalue [code])
      break ;
  }
  extra = extrasDistance [code] ;
  value = distance - bases [code] ;
  return ;
}

  // extras [n] is the number of extra bits for code n.
  static const UBYTE1 extrasLength [PngMaxLengthCodes - PngFirstLengthCode] =
    {
      0, 0, 0, 0, 0, 0, 0, 0,
      1, 1, 1, 1, 2, 2, 2, 2,
      3, 3, 3, 3, 4, 4, 4, 4,
      5, 5, 5, 5, 0, 0, 0,
    } ;

inline static void LengthToCode (unsigned int length, unsigned int &code,
                                 unsigned int &extra, unsigned int &value)
{
  // codes [n] is the length for for length n - 3.
  static const UBYTE2 codes [PngLongestLength-2] =
    {
      257, 258, 259, 260, 261, 262, 263, 264,
      265, 265, 266, 266, 267, 267, 268, 268,
      269, 269, 269, 269, 270, 270, 270, 270,
      271, 271, 271, 271, 272, 272, 272, 272,
      273, 273, 273, 273, 273, 273, 273, 273,
      274, 274, 274, 274, 274, 274, 274, 274,
      275, 275, 275, 275, 275, 275, 275, 275,
      276, 276, 276, 276, 276, 276, 276, 276,
      277, 277, 277, 277, 277, 277, 277, 277,
      277, 277, 277, 277, 277, 277, 277, 277,
      278, 278, 278, 278, 278, 278, 278, 278,
      278, 278, 278, 278, 278, 278, 278, 278,
      279, 279, 279, 279, 279, 279, 279, 279,
      279, 279, 279, 279, 279, 279, 279, 279,
      280, 280, 280, 280, 280, 280, 280, 280,
      280, 280, 280, 280, 280, 280, 280, 280,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      281, 281, 281, 281, 281, 281, 281, 281,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      282, 282, 282, 282, 282, 282, 282, 282,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      283, 283, 283, 283, 283, 283, 283, 283,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 284,
      284, 284, 284, 284, 284, 284, 284, 285,
    } ;

   static const UBYTE2 basevalues [PngMaxLengthCodes - PngFirstLengthCode] =
    {
        3,   4,   5,   6,   7,   8,   9,  10,
       11,  13,  15,  17,  19,  23,  27,  31,
       35,  43,  51,  59,  67,  83,  99, 115,
      131, 163, 195, 227, 258,   0,   0,
    } ;

  code = codes [length - 3] ;
  extra = extrasLength [code - PngFirstLengthCode] ;
  value = length - basevalues [code - PngFirstLengthCode] ;
  return ;
}

//
//  Description:
//
//    Class Default Constructor
//
PngEncoder::PngEncoder ()
{
  Initialize () ;
  return ;
}

//
//  Description:
//
//    Class Copy Constructor
//
PngEncoder::PngEncoder (const PngEncoder &source)
{
  Initialize () ;
  DoCopy (source) ;
  return ;
}

//
//  Description:
//
//    Class Destructor
//
PngEncoder::~PngEncoder ()
{
  FreeBuffers () ;
  delete distance_table ;
  delete length_table ;
  delete length_length_table ;
  return ;
}

//
//  Descriptor:
//
//    Assignment Operator
//
//  Parameters:
//    source: The object to copy
//
PngEncoder &PngEncoder::operator=(const PngEncoder &source)
{
  DoCopy (source) ;
  return *this ;
}

//
//  Description:
//
//    Common class initialization function. This function is
//    called from constructors to initialize the object.
//
void PngEncoder::Initialize ()
{ countZamina=0;
  countZaminaV=0;
  countZaminaG=0;
  compressionVertical=true;
  compressionGorizontal=true;
  MakeCrcTable () ;
  software_string = "Colosseum Builders Image Library" ;
  trebaOptimizeLZBlock = true; //   LZ- 
  current_image = NULL ;
  compression_level = DefaultCompression ;

  distance_table = new PngHuffmanEncoder ;
  length_table = new PngHuffmanEncoder ;
  length_length_table = new PngHuffmanEncoder ;

  block_buffer = NULL ;
  trebaZaminaBuffer = NULL ;
  pozAnalizZamina = NULL;
  pozImageBuffer = NULL;
  block_buffer_size = maxBlockBufferSize;
  return ;
}

//
//  Description:
//
//    Class copy function. This function is called from the copy
//    constructor and assignment operator.
//
//  Parameters:
//    source:  The object to copy
//
void PngEncoder::DoCopy (const PngEncoder &source)
{
  title_string = source.title_string ;
  author_string = source.author_string ;
  description_string = source.description_string ;
  copyright_string = source.copyright_string ;
  software_string = source.software_string ;
  disclaimer_string = source.disclaimer_string ;
  warning_string = source.warning_string ;
  source_string = source.source_string ;
  comment_string = source.comment_string ;
  block_buffer_size = source.block_buffer_size ;
  BitmapImageEncoder::DoCopy (source) ;
  return ;
}

//
//  Description:
//
//    This function writes an image to a PNG stream.
//
//  Parameters:
//    strm:  The output stream
//    image:  The image to outpu
//
void PngEncoder::WriteImage (std::ostream &strm, BitmapImage &image)
{
  current_image = &image ;
  output_stream = &strm ;

  try
  {
    DoWrite () ;
  }
  catch (...)
  {
    FreeBuffers () ;
    throw ;
  }

  FreeBuffers () ;
  return ;
}

//
//  Description:
//
//    This function frees the buffers allowed during the encoding
//    process.
//
void PngEncoder::FreeBuffers ()
{
  current_image = NULL ;
  output_stream = NULL ;
  delete [] hash_values ; hash_values = NULL ;
  delete [] hash_table ; hash_table = NULL ;
  delete [] output_buffer ; output_buffer = NULL ;
  delete [] imageData; imageData = NULL;
  delete [] countHashValue; countHashValue=NULL;
  if (compressionVertical)
   {delete [] vhash_values ; vhash_values = NULL ;
    delete [] vhash_table ; vhash_table = NULL ;
    delete [] vimageData; vimageData = NULL;
    delete [] vcountHashValue; vcountHashValue=NULL;
   }
  if (compressionGorizontal)
   {delete [] gimageData; gimageData = NULL;
    delete [] ghash_values ; ghash_values = NULL ;
    delete [] ghash_table ; ghash_table = NULL ;
    delete [] gcountHashValue; gcountHashValue=NULL;
   }
  delete [] dimageData; dimageData = NULL; 
  return ;
}

//
//  Description:
//
//    This function calculates the hash value from a location
//    in the input stream. We create the hash value by
//    extracting a fixed number of the low order bits from each
//    of the next three data values.
//
//  Parameters:
//    index: The index into the lookahead buffer
//
//  Return Value:
//    The hash value
//
unsigned int PngEncoder::HashValue (unsigned long index)
{
  return Hash (imageData[index],
               imageData[index+1],
               imageData[index+2]) ;
}

unsigned int PngEncoder::VHashValue (unsigned long index)
{
  return Hash (vimageData[index],
               vimageData[index+1],
               vimageData[index+2]) ;
}

unsigned int PngEncoder::GHashValue (unsigned long index)
{
  return Hash (gimageData[index],
               gimageData[index+1],
               gimageData[index+2]) ;
}
//
//  Description:
//
//    This function finds the longest string in the LZ
//    window that matches the data in the lookahead buffer.
//
//    Since the hash changes can get very long under certain
//    circumstances we use the search_limit to set a limit
//    to how far we search in the tree.
//
//  Parameters:
//    bestlength: The longest match found (0=>No match)
//    bestoffset: The location where the best match was found
//

void PngEncoder::LongestMatch (unsigned int &bestlength,
                               unsigned int &bestoffset)
 {bestlength = 0;
  unsigned long bestPoz, lenAnalizHash=lenKeyHash;
  bool znadeno;
  //  -   
  unsigned int hashvalue = HashValue (currentPozImage);
  if (hash_table [hashvalue] == NULL)
    return ;

  HashEntry *current;
  unsigned int countRivni, delta=0, countPoz=0;
  long countZaluchokPoz;
  current=hash_table[hashvalue];
  countZaluchokPoz=countHashValue[hashvalue];
startAnalizPoz:
  while (countPoz++<search_limit && countZaluchokPoz>0 &&
         currentPozImage-(current->index-delta)<=PngWindowSize)
   {compareStartPoslidovnist(currentPozImage, current->index-delta, countRivni, bestlength);
    countZaluchokPoz--;
    if (countRivni>bestlength)
     {bestlength=countRivni;
      bestPoz=current->index-delta;
      if (bestlength==PngLongestLength) //  
       break;
      if (countZaluchokPoz>bestlength-lenAnalizHash) //     
       {znadeno=false;
        while (lenAnalizHash<=bestlength)
         {lenAnalizHash+=3;
          unsigned int newHashValue = HashValue (currentPozImage+lenAnalizHash-lenKeyHash);
          if (countZaluchokPoz>countHashValue[newHashValue])
           {countZaluchokPoz=countHashValue[newHashValue];
            current=hash_table[newHashValue];
            delta=lenAnalizHash-lenKeyHash;
            znadeno=true;
           }
         }
        if (znadeno)
         goto startAnalizPoz;
       }
     }
    current=current->next;
   }
  if (currentPozImage+bestlength >= countByteImage)
   bestlength=countByteImage-currentPozImage;
  if (bestlength<minLenZamina)
   {bestlength=0;
    return;
   }
  bestoffset=(currentPozImage-bestPoz)/3;
  return ;
}

void PngEncoder::VLongestMatch (unsigned int &bestlength,
                                unsigned int &bestoffset)
 {bestlength = 0;
  unsigned long bestPoz, lenAnalizHash=lenKeyHash;
  bool znadeno;
  //  -   
  unsigned int hashvalue = VHashValue (currentPozImage);
  if (vhash_table [hashvalue] == NULL)
    return ;

  HashEntry *current;
  unsigned int countRivni, delta=0, countPoz=0;
  long countZaluchokPoz;
  current=vhash_table[hashvalue];
  countZaluchokPoz=vcountHashValue[hashvalue];
startAnalizPoz:
  while (countPoz++<search_limit && countZaluchokPoz>0 &&
         currentPozImage-(current->index-delta)<=PngWindowSize)
   {vcompareStartPoslidovnist(currentPozImage, current->index-delta, countRivni, bestlength);
    countZaluchokPoz--;
    if (countRivni>bestlength)
     {bestlength=countRivni;
      bestPoz=current->index-delta;
      if (bestlength==PngLongestLength) //  
       break;
      if (countZaluchokPoz>bestlength-lenAnalizHash) //     
       {znadeno=false;
        while (lenAnalizHash<=bestlength)
         {lenAnalizHash+=3;
          unsigned int newHashValue = VHashValue (currentPozImage+lenAnalizHash-lenKeyHash);
          if (countZaluchokPoz>vcountHashValue[newHashValue])
           {countZaluchokPoz=vcountHashValue[newHashValue];
            current=vhash_table[newHashValue];
            delta=lenAnalizHash-lenKeyHash;
            znadeno=true;
           }
         }
        if (znadeno)
         goto startAnalizPoz;
       }
     }
    current=current->next;
   }
  if (currentPozImage+bestlength >= countByteImage)
   bestlength=countByteImage-currentPozImage;
  if (bestlength<minLenZamina)
   {bestlength=0;
    return;
   }
  bestoffset=(currentPozImage-bestPoz)/3;
  return ;
}

void PngEncoder::GLongestMatch (unsigned int &bestlength,
                                unsigned int &bestoffset)
 {bestlength = 0;
  unsigned long bestPoz, lenAnalizHash=lenKeyHash;
  bool znadeno;
  //  -   
  unsigned int hashvalue = GHashValue (currentPozImage);
  if (ghash_table [hashvalue] == NULL)
    return ;

  HashEntry *current;
  unsigned int countRivni, delta=0, countPoz=0;
  long countZaluchokPoz;
  current=ghash_table[hashvalue];
  countZaluchokPoz=gcountHashValue[hashvalue];
startAnalizPoz:
  while (countPoz++<search_limit && countZaluchokPoz>0 &&
         currentPozImage-(current->index-delta)<=PngWindowSize)
   {gcompareStartPoslidovnist(currentPozImage, current->index-delta, countRivni, bestlength);
    countZaluchokPoz--;
    if (countRivni>bestlength)
     {bestlength=countRivni;
      bestPoz=current->index-delta;
      if (bestlength==PngLongestLength) //  
       break;
      if (countZaluchokPoz>bestlength-lenAnalizHash) //     
       {znadeno=false;
        while (lenAnalizHash<=bestlength)
         {lenAnalizHash+=3;
          unsigned int newHashValue = GHashValue (currentPozImage+lenAnalizHash-lenKeyHash);
          if (countZaluchokPoz>gcountHashValue[newHashValue])
           {countZaluchokPoz=gcountHashValue[newHashValue];
            current=ghash_table[newHashValue];
            delta=lenAnalizHash-lenKeyHash;
            znadeno=true;
           }
         }
        if (znadeno)
         goto startAnalizPoz;
       }
     }
    current=current->next;
   }
  if (currentPozImage+bestlength >= countByteImage)
   bestlength=countByteImage-currentPozImage;
  if (bestlength<minLenZamina)
   {bestlength=0;
    return;
   }
  bestoffset=(currentPozImage-bestPoz)/3;
  return ;
}

void PngEncoder::compareStartPoslidovnist(long adres1,
                                          long adres2,
                                          unsigned int &countRivni,
                                          unsigned int bestLength)
 {countRivni=0;
  if (adres1<1 || adres2<0)
   return;
  if (imageData[adres1+bestLength]!=imageData[adres2+bestLength] ||
      imageData[adres1+bestLength+1]!=imageData[adres2+bestLength+1] ||
      imageData[adres1+bestLength+2]!=imageData[adres2+bestLength+2]) //   
   return;
  for ( ; countRivni<bestLength; countRivni+=3)
   if (imageData[adres1+countRivni]!=imageData[adres2+countRivni] ||
       imageData[adres1+countRivni+1]!=imageData[adres2+countRivni+1] ||
       imageData[adres1+countRivni+2]!=imageData[adres2+countRivni+2])
    return;
  countRivni+=3;
  for ( ; countRivni<PngLongestLength; countRivni+=3)
   if (imageData[adres1+countRivni]!=imageData[adres2+countRivni] ||
       imageData[adres1+countRivni+1]!=imageData[adres2+countRivni+1] ||
       imageData[adres1+countRivni+2]!=imageData[adres2+countRivni+2])
    return;
  return;
 }

void PngEncoder::vcompareStartPoslidovnist(long adres1,
                                           long adres2,
                                           unsigned int &countRivni,
                                           unsigned int bestLength)
 {countRivni=0;
  if (adres1<1 || adres2<0)
   return;
  if (vimageData[adres1+bestLength]!=vimageData[adres2+bestLength] ||
      vimageData[adres1+bestLength+1]!=vimageData[adres2+bestLength+1] ||
      vimageData[adres1+bestLength+2]!=vimageData[adres2+bestLength+2]) //   
   return;
  for ( ; countRivni<bestLength; countRivni+=3)
   if (vimageData[adres1+countRivni]!=vimageData[adres2+countRivni] ||
       vimageData[adres1+countRivni+1]!=vimageData[adres2+countRivni+1] ||
       vimageData[adres1+countRivni+2]!=vimageData[adres2+countRivni+2])
    return;
  countRivni+=3;
  for ( ; countRivni<PngLongestLength; countRivni+=3)
   if (vimageData[adres1+countRivni]!=vimageData[adres2+countRivni] ||
       vimageData[adres1+countRivni+1]!=vimageData[adres2+countRivni+1] ||
       vimageData[adres1+countRivni+2]!=vimageData[adres2+countRivni+2])
    return;
  return;
 }

void PngEncoder::gcompareStartPoslidovnist(long adres1,
                                           long adres2,
                                           unsigned int &countRivni,
                                           unsigned int bestLength)
 {countRivni=0;
  if (adres1<1 || adres2<0)
   return;
  if (gimageData[adres1+bestLength]!=gimageData[adres2+bestLength] ||
      gimageData[adres1+bestLength+1]!=gimageData[adres2+bestLength+1] ||
      gimageData[adres1+bestLength+2]!=gimageData[adres2+bestLength+2]) //   
   return;
  for ( ; countRivni<bestLength; countRivni+=3)
   if (gimageData[adres1+countRivni]!=gimageData[adres2+countRivni] ||
       gimageData[adres1+countRivni+1]!=gimageData[adres2+countRivni+1] ||
       gimageData[adres1+countRivni+2]!=gimageData[adres2+countRivni+2])
    return;
  countRivni+=3;
  for ( ; countRivni<PngLongestLength; countRivni+=3)
   if (gimageData[adres1+countRivni]!=gimageData[adres2+countRivni] ||
       gimageData[adres1+countRivni+1]!=gimageData[adres2+countRivni+1] ||
       gimageData[adres1+countRivni+2]!=gimageData[adres2+countRivni+2])
    return;
  return;
 }

//    
void PngEncoder::FreePozHashEntry(unsigned long entry)
 {if (entry>=PngWindowSize)
   {unsigned long indexValue=entry % PngWindowSize; //   -
    unsigned int oldHashValue=HashValue(hash_values[indexValue].index);
    if (countHashValue[oldHashValue]==1) //    
     {countHashValue[oldHashValue]=0;
      return;
     }
    countHashValue[oldHashValue]--;
   }
 }

//   hash_table [hashvalue]  entry
void PngEncoder::MoveHashEntry (unsigned long entry, unsigned int hashvalue)
{ //   -
  unsigned long indexValue=entry % PngWindowSize; //   -
  hash_values[indexValue].next=hash_table[hashvalue];
  hash_values[indexValue].index=entry;
  hash_table[hashvalue]=&hash_values[indexValue];
  countHashValue[hashvalue]++;
}

//    
void PngEncoder::VFreePozHashEntry(unsigned long entry)
 {if (entry>=PngWindowSize)
   {unsigned long indexValue=entry % PngWindowSize; //   -
    unsigned int oldHashValue=VHashValue(vhash_values[indexValue].index);
    if (vcountHashValue[oldHashValue]==1) //    
     {vcountHashValue[oldHashValue]=0;
      return;
     }
    vcountHashValue[oldHashValue]--;
   }
 }

//   hash_table [hashvalue]  entry
void PngEncoder::VMoveHashEntry (unsigned long entry, unsigned int hashvalue)
{ //   -
  unsigned long indexValue=entry % PngWindowSize; //   -
  vhash_values[indexValue].next=vhash_table[hashvalue];
  vhash_values[indexValue].index=entry;
  vhash_table[hashvalue]=&vhash_values[indexValue];
  vcountHashValue[hashvalue]++;
}

//    
void PngEncoder::GFreePozHashEntry(unsigned long entry)
 {if (entry>=PngWindowSize)
   {unsigned long indexValue=entry % PngWindowSize; //   -
    unsigned int oldHashValue=GHashValue(ghash_values[indexValue].index);
    if (gcountHashValue[oldHashValue]==1) //    
     {gcountHashValue[oldHashValue]=0;
      return;
     }
    gcountHashValue[oldHashValue]--;
   }
 }

//   hash_table [hashvalue]  entry
void PngEncoder::GMoveHashEntry (unsigned long entry, unsigned int hashvalue)
{ //   -
  unsigned long indexValue=entry % PngWindowSize; //   -
  ghash_values[indexValue].next=ghash_table[hashvalue];
  ghash_values[indexValue].index=entry;
  ghash_table[hashvalue]=&ghash_values[indexValue];
  gcountHashValue[hashvalue]++;
}

//
//  Description:
//
//    This function finds the length encoding for
//    a length or distance table.
//
//  Parameters:
//    function: The function to process the code
//    lengths: lengths [n] = the length of the huffman code for n.
//    count: The number of coe lengths
//
void PngEncoder::FindLengthCodes (LENGTHFUNCTION function,
                                  UBYTE1 lengths [],
                                  unsigned int count)
{
  for (unsigned int ii = 0, jj ; ii < count ; ii = jj)
  {
    if (lengths [ii] != 0)
    {
      (this->*function) (ii, lengths [ii], 0, 0) ;
      jj = ii + 1 ;
    }
    else
    {
      // Attempt to compact runs of zero length codes.
      // First find the number of consecutive zeros.
      for (jj = ii + 1;
           lengths [jj] == lengths [jj-1]
           && jj < count ;
           ++ jj)
      {
      }

      // We need at least 3 consecutive zeros to compact them.
      switch (jj - ii)
      {
      case 1:
        (this->*function) (ii, lengths [ii], 0, 0) ;
        break ;
      case 2:
        (this->*function) (ii, lengths [ii], 0, 0) ;
        (this->*function) (ii + 1, lengths [ii], 0, 0) ;
        break ;
      default:
        {
          // We have at least three zeros.

          int kk = jj - ii ;
          if (kk > 138)
          {
            kk = 138 ;
            jj = ii + kk ;
          }
          if (kk > 10)
          {
            (this->*function) (ii, 18, 7, kk - 11) ;
          }
          else
          {
            (this->*function) (ii, 17, 3, kk - 3) ;
          }
        }
        break ;
      }
    }
  }
  return ;
}


//
//  Description:
//
//    This function is passed as a parameter to FindLengthCodes.
//    It is use to find the frequency for each code.
//
//  Parameters:
//    code:  The code generated by FindLengthCodes
//
void PngEncoder::GatherLengthCounts (unsigned int,
                                     unsigned int code,
                                     unsigned int,
                                     unsigned int)
{
  length_length_table->IncrementFrequency (code) ;
  return ;
}

//
//  Description:
//
//    This function is passed as a parameter to FindLengthCodes.
//    It is use to encode and output the code to the output stream.
//
//  Parameters:
//    code:  The code generated by FindLengthCodes
//    extra:  The number of extra bits of data for the code
//    value:  The extra value
//
void PngEncoder::OutputLengthCounts (unsigned int,   // Index - Not used here
                                     unsigned int code,
                                     unsigned int extra,
                                     unsigned int value)
{
  UBYTE2 huffmancode ;
  UBYTE1 huffmansize ;
  length_length_table->Encode (code, huffmancode, huffmansize) ;

  OutputDataBits (huffmancode, huffmansize) ;
  if (extra != 0)
    OutputDataBits (value, extra) ;

  return ;
}

//
//  Description:
//
//    This function is called to start a new chunk.
//
//  Parameters:
//    type:  The PNG Chunk type (e.g. "IHDR", "IDAT")
//
void PngEncoder::StartChunk (const char type [5])
{
  chunk_type = ChunkType (type) ;
  bit_position = 0 ;
  byte_position = 0 ;
  output_buffer [0] = 0 ;
  return ;
}

//
//  Description:
//
//    This function outputs the current chunk.
//
void PngEncoder::WriteCurrentChunk ()
{
  // Calculate the CRC value for the chunk type and data
  UBYTE4 crc = 0xFFFFFFFFL ;
  crc = CRC32 (crc, &chunk_type, sizeof (chunk_type)) ;
  crc = CRC32 (crc, output_buffer, byte_position) ;
  crc ^= 0xFFFFFFFFL ;
  crc = SystemToBigEndian (crc) ;
  // Output the chunk. Length, type, data, CRC
  UBYTE4 length = SystemToBigEndian (byte_position) ;
  output_stream->write ((char *) &length, sizeof (length)) ;
  output_stream->write ((char *) &chunk_type, sizeof (chunk_type)) ;
  output_stream->write ((char *) output_buffer, byte_position) ;
  output_stream->write ((char *) &crc, sizeof (crc)) ;

  byte_position = 0 ;
  bit_position = 0 ;
  return ;
}

//
//  Description:
//
//    This function writes a block of data to the current chunk.
//
//  Parameters:
//    data: Pointer to the data to write
//    length: Number of bytes to write
//
void PngEncoder::OutputBlock (const void *data, unsigned int length)
{
  if (byte_position + length > OutputBufferSize)
    throw EPngError ("Output Buffer Overrun") ;

  memcpy (&output_buffer [byte_position], data, length) ;
  byte_position += length ;
  return;
}

//
//  Description:
//
//    This function writes the Adler value calculated for
//    the input data to the current chunk.
//    
void PngEncoder::OutputAdler ()
{
  FlushBitBuffer () ;
  adler_value = SystemToBigEndian (adler_value) ;
  UBYTE1 *buffer = (UBYTE1*) &adler_value ;
  for (unsigned int ii = 0 ; ii < sizeof (adler_value) ; ++ ii)
  {
    if (byte_position == OutputBufferSize)
      WriteCurrentChunk () ;

    output_buffer [byte_position] = buffer [ii] ;
    ++ byte_position ;
  }
  return ;
}

//
//  Description:
//
//    This function writes a number of data bits within an IDAT block.
//
//  Parameters:
//    data: The bits to output
//    length:  The number of bits to output
//
void PngEncoder::OutputDataBits (unsigned int data, unsigned int length)
{
  for (unsigned int ii = 0 ; ii < length ; ++ ii)
  {
    if (bit_position >= CHAR_BIT)
      FlushBitBuffer () ;

    if ((data & (1 << ii)) != 0)
    {
      output_buffer [byte_position] |= (1 << bit_position) ;
    }
    ++ bit_position ;
  }
  return ;
}

//
//  Description:
//
//    This function flushes any partial bytes that have been
//    written. This allows us to start writing byte data or 
//    start the next data byte.
//
void PngEncoder::FlushBitBuffer ()
{
  if (bit_position == 0)
    return ;

  ++ byte_position ;
  if (byte_position >= OutputBufferSize)
  {
    WriteCurrentChunk () ;
    StartChunk ("IDAT") ;
  }
  output_buffer [byte_position] = 0 ;
  bit_position = 0 ;
  return ;
}

//
//  Description:
//
//    This function encodes the image data. Depending upon
//    the function arguments this function either gathers
//    Huffman usage statistics or Huffman-encodes the data.
//
//  Return Value:
//    true => All data in the image has been processed
//    false => More data remains
//
bool PngEncoder::ProcessImageData ()
{
  // Here we loop until we either fill the block_buffer or
  // reach the end of the image.
  for (block_buffer_count = 0 ;
       block_buffer_count < block_buffer_size - 2
       && currentPozImage < countByteImage ;
       ++ block_buffer_count, ++currentGlobalPoz)
   {unsigned int length, offset, zmLen;
    if (globalBlockBuffer[currentGlobalPoz]==0)
     length=0;
    else
     {zmLen=(globalBlockBuffer[currentGlobalPoz]-256)%3;
      length=globalBlockBuffer[currentGlobalPoz]-256-zmLen;
      currentGlobalPoz++;
      offset=globalBlockBuffer[currentGlobalPoz];
     }
    if (length == 0)
     {UBYTE1 literal = indexPointPaletr[currentPozImage/3];
      block_buffer [block_buffer_count] = literal ;
      length_table->IncrementFrequency (literal) ;
      adler_value = Adler (adler_value,imageData[currentPozImage]);
      currentPozImage++;
      adler_value = Adler (adler_value,imageData[currentPozImage]);
      currentPozImage++;
      adler_value = Adler (adler_value,imageData[currentPozImage]);
      currentPozImage++;
     }
    else
     {// Gather Huffman Statistics
      unsigned int code, codeD ;
      unsigned int extra ;
      unsigned int value ;
      LengthToCode (length+zmLen, code, extra, value) ;
	   length_table->IncrementFrequency (code) ;
      DistanceToCode (offset, codeD, extra, value) ;
	   distance_table->IncrementFrequency (codeD) ;
      if (trebaOptimizeLZBlock)
       {trebaZaminaBuffer[block_buffer_count]=1;  //     
        if (length+zmLen<minLenZaminaLiteral) //       ,   
         {pozImageBuffer[block_buffer_count]=currentPozImage/3;
          pozAnalizZamina[countAnalizZamina]=block_buffer_count;    //     
          countAnalizZamina++;
         }
       }
      switch (zmLen)
       {case 0:countZamina++;
               break;
        case 1:countZaminaV++;
               break;
        case 2:countZaminaG++;
       }
      // We have found a match.  First update the hash table and
      // copy the data in the LZ window.
      for (unsigned int ii = 0 ; ii < length ; ii++)
       {adler_value = Adler (adler_value,imageData[currentPozImage]);
        currentPozImage++;
       }
      block_buffer [block_buffer_count] = 256 + length + zmLen;
      ++ block_buffer_count ;
      block_buffer [block_buffer_count] = offset ;
     }
   }
  length_table->IncrementFrequency (PngEndCode) ;
  if (currentPozImage == countByteImage)
    return true ;
  else
    return false ;
}

//
//  Description:
//
//    This function outputs a single data byte.
//
//  Parameters:
//    value: The value to output
//
void PngEncoder::OutputByte (UBYTE1 value)
{
  output_buffer [byte_position] = value ;
  ++ byte_position ;
  return ;
}

//
//  Description:
//
//    This function calls the progress function if it has been
//    defined.
//
//  Parameters:
//    percent:  The completion percentage (0..100)
//
void PngEncoder::CallProgressFunction (unsigned int percent)
{
  if (progress_function == NULL)
    return ;

  if (percent > 100)
    percent = 100 ;

  bool cancel = false ;
  progress_function (*this, progress_data, 1, 1,
                     percent, cancel) ;
  return ;
}

//
//  Description:
// 
//    This function creates a tEXt chunk.
//
//  Parameters:
//    keyword:  The chunk keyword
//    value:  The keyword value
//
void PngEncoder::WriteText (const std::string &keyword,
                            const std::string &value)
{
  if (keyword.length () > 79)
    throw EPngError ("tEXt Keyword Too Long") ;
  if (value.length () + keyword.length () + 1 > OutputBufferSize)
    throw EPngError ("tEXt Value Too Long") ;

  StartChunk ("tEXt") ;
  OutputBlock (keyword.c_str (), keyword.length ()) ;
  OutputByte (0) ;
  OutputBlock (value.c_str (), value.length ()) ;
  WriteCurrentChunk () ;
  return ;
}

//
//  Description:
//
//    This function outputs tEXt blocks for any keywords
//    that have values assigned to them.
//
void PngEncoder::WriteTextBlocks ()
{
  if (title_string != "")
    WriteText ("Title", title_string) ;
  if (author_string != "")
    WriteText ("Author", author_string) ;
  if (description_string != "")
    WriteText ("Description", description_string) ;
  if (copyright_string != "")
    WriteText ("Copyright", copyright_string) ;
  if (software_string != "")
    WriteText ("Software", software_string) ;
  if (disclaimer_string != "")
    WriteText ("Disclaimer", disclaimer_string) ;
  if (warning_string != "")
    WriteText ("Warning", warning_string) ;
  if (source_string != "")
    WriteText ("Source", source_string) ;
  if (comment_string != "")
    WriteText ("Comment", comment_string) ;
  return ;
}

//
//  Description:
//
//    This function sets the compression level used.  The compression level
//    determines the depth to which hash chains are searched.
//
//  Parameters:
//    value:  The new compression level
//
void PngEncoder::SetCompressionLevel (PngEncoder::CompressionLevel value)
{
  if (value < 0 || value > 3)
    throw EPngError ("Invalid Compression Level") ;

  compression_level = value ;
  return ;
}

//
//  Description:
//
//    The function returns the current compression level.
//
//  Return Value:
//    The compresion level.
//
PngEncoder::CompressionLevel PngEncoder::GetCompressionLevel () const
{
  return compression_level ;
}

void PngEncoder::SetCompressionVertical ()
{
  compressionVertical=false ;
}

void PngEncoder::SetCompressionGorizontal ()
{
  compressionGorizontal=false ;
}

//
//  Description:
//
//    This function performs the actual output of the image.
//

UBYTE4 sizeEntropiCode(UBYTE4 *masFreq)
 {UBYTE4 count=0, i;

  double size=0;

  for (i=0; i<256; i++)

   {count+=masFreq[i];

    if (masFreq[i]>1)

     size+=masFreq[i]*log(masFreq[i]);

   }

  if (count)

   size=(count*log(count)-size)/log(2);

  return size;

 }

void PngEncoder::initCM() //       BRG
 {//   
  UBYTE2 i, j, m, n, r;
  //  
  for (j=0; j<image_row;j++)
   for (i=0; i<image_col;i++)
    {imageData[j*row_width+i*countComponent]=(*current_image)[j][i*countComponent];
     imageData[j*row_width+i*countComponent+1]=(*current_image)[j][i*countComponent+1];
     imageData[j*row_width+i*countComponent+2]=(*current_image)[j][i*countComponent+2];
    }
  // LeftPredict
  UBYTE4 k;
  for (k=(UBYTE4)image_row*row_width-1; k>2; k--)
   imageData[k]-=imageData[k-3];
  UBYTE4 freq[256];
  UBYTE4 prognozLen[3][3], minusLen;
  for (m=0; m<=2; m++) //  
   for (n=0; n<=2; n++) //   
    {memset(freq, 0, sizeof(freq));
     if (m==n) //     0
      {for (i = 0 ; i < image_row; i++)
        for (j = 0; j<row_width; j+=3)
         freq[imageData[i*row_width+j+m]]++;
       prognozLen[m][n]=sizeEntropiCode(freq);
      }
     else
      {if (n>m)
        {for (i = 0 ; i < image_row; i++)
          for (j = 0; j<row_width; j+=3)
           freq[(UBYTE1)(imageData[i*row_width+j+m]-imageData[i*row_width+j+n])]++;
         prognozLen[m][n]=sizeEntropiCode(freq);
        }
       else
        prognozLen[m][n]=prognozLen[n][m];
      }
    }
  UBYTE1 c11, c12, c21, c22;
  cm11=0, cm12=0, cm21=0, cm22=0;
  minusLen=0;
  for (c11=0; c11<=2; c11++)
   for (c12=0; c12<=2; c12++)
    if (c11!=c12 && prognozLen[c11][c12]<prognozLen[c11][c11]) //     
     {if (prognozLen[c11][c11]-prognozLen[c11][c12]>minusLen)
       {minusLen=prognozLen[c11][c11]-prognozLen[c11][c12];
        cm11=c11;
        cm12=c12;
        cm21=0;
        cm22=0;
       }
      for (c21=c11+1; c21<=2; c21++)
       for (c22=0; c22<=2; c22++)
        if (c21!=c22 && (c21!=c12 || c22!=c11) && prognozLen[c21][c22]<prognozLen[c21][c21]) //     
         if (prognozLen[c11][c11]-prognozLen[c11][c12]+
             prognozLen[c21][c21]-prognozLen[c21][c22]>minusLen)
          {minusLen=prognozLen[c11][c11]-prognozLen[c11][c12]+
                    prognozLen[c21][c21]-prognozLen[c21][c22];
           cm11=c11;
           cm12=c12;
           cm21=c21;
           cm22=c22;
          }
     }
  if (cm11==cm22) //   
   {i=cm11;
    cm11=cm21;
    cm21=i;
    i=cm12;
    cm12=cm22;
    cm22=i;
   }
  if (cm11!=cm12) //  
   {for (i = 0 ; i < image_row; i++)
     for (r = 0; r<row_width; r+=3)
      (*current_image) [i][r+cm11]-=(*current_image)[i][r+cm12];
   }
  if (cm21!=cm22) //  
   {for (i = 0 ; i < image_row; i++)
     for (r = 0; r<row_width; r+=3)
      (*current_image) [i][r+cm21]-=(*current_image)[i][r+cm22];
   }
 }

void PngEncoder::writeCM()
 {if (cm11!=cm12)
   {StartChunk ("cmOd");
    UBYTE1 cm=cm11<<6 | cm12<<4 | cm21<<2 |cm22;
    OutputBlock (&cm, 1) ;
    WriteCurrentChunk () ;
   }
 }

void PngEncoder::writeNotCP()
 {if (!trebaCorectPredict)
   {StartChunk ("ncPr");
    WriteCurrentChunk () ;
   }
 }

unsigned char MedPredict(unsigned char Rb, unsigned char Ra, unsigned char Rc)
 {register unsigned char minx, maxx, Px;
  if (Rb > Ra)
   {minx = Ra;
    maxx = Rb;
   }
  else
   {maxx = Ra;
    minx = Rb;
   }
  if (Rc >= maxx)
   Px = minx;
  else
   if (Rc <= minx)
    Px = maxx;
   else
    Px = Ra + Rb - Rc;
  return Px;
 }

void PngEncoder::DoWrite ()
{
  // Needed because MSVC++ does not follow standard scoping rules
  // in for statements.
  unsigned int ii ;

  if (current_image->BitCount () == 24)
  {
    row_width = 3 * current_image->Width () ;
    countComponent = 3 ;
  }
  else
  {
    row_width = (current_image->BitCount () * current_image->Width ()
                 + CHAR_BIT - 1) / CHAR_BIT ;
    countComponent = 1 ;
  }

  // Convert the compression level to the maximum depth hash
  // chains are searched.
  switch (compression_level)
  {
  case FastestCompression:
    search_limit = 0 ;
    break;
  case FastCompression:
    search_limit = 64 ;
    break ;
  case DefaultCompression:
    search_limit = 128 ;
    break ;
  case MaximumCompression:
    search_limit = UINT_MAX ;
    break ;
  default:
    throw EPngError ("Invalid Compression Level") ;
  }

  image_row=current_image->Height();
  image_col=current_image->Width();
  imageData=new UBYTE1[image_row*row_width+500];
  memset(imageData+image_row*row_width, 0, 500);
  initCM(); //    

  //      
  unsigned int i,j;
  unsigned int maxx=current_image->BytesPerRow();
  for (j=0; j<image_row;j++)
   for (i=0; i<image_col;i++)
    {imageData[j*row_width+i*countComponent]=current_image->ImageData()[(image_row-j-1)*maxx+i*countComponent+2];
     imageData[j*row_width+i*countComponent+1]=current_image->ImageData()[(image_row-j-1)*maxx+i*countComponent+1];
     imageData[j*row_width+i*countComponent+2]=current_image->ImageData()[(image_row-j-1)*maxx+i*countComponent];
    }
  PngWindowSize= (1<<15)*3; // -  PNG-
  PngWindowSize-=PngWindowSize % 3;

  hash_values = new HashEntry [PngWindowSize] ;
  hash_table = new PHashEntry[HashTableSize] ;
  countHashValue = new unsigned long[HashTableSize];

  if (compressionVertical)
   {vimageData=new UBYTE1[image_row*row_width+500];
    memset(vimageData+image_row*row_width, 0, 500);
    for (i=0; i<row_width;i++)
     vimageData[i]=imageData[i];
    for (j=1; j<image_row;j++)
     for (i=0; i<row_width;i++)
      vimageData[j*row_width+i]=imageData[j*row_width+i]-imageData[(j-1)*row_width+i];
    vhash_values = new HashEntry [PngWindowSize] ;
    vhash_table = new PHashEntry[HashTableSize] ;
    vcountHashValue = new unsigned long[HashTableSize];
   }

  if (compressionGorizontal)
   {gimageData=new UBYTE1[image_row*row_width+500];
    memset(gimageData+image_row*row_width, 0, 500);
    for (i=0; i<countComponent;i++)
     gimageData[i]=imageData[i];
    for (i=countComponent; i<image_row*row_width; i++)
     gimageData[i]=imageData[i]-imageData[i-countComponent];
    ghash_values = new HashEntry [PngWindowSize] ;
    ghash_table = new PHashEntry[HashTableSize] ;
    gcountHashValue = new unsigned long[HashTableSize];
   }

  dimageData=new UBYTE1[image_row*row_width];
  for (i=0; i<countComponent;i++)
   {dimageData[i]=imageData[i]-corectComponent[i].JPEGLSPredict(0, 0, 0, 0, imageData[i]);;
    for (j=countComponent+i; j<row_width-countComponent; j+=countComponent)
     dimageData[j]=imageData[j]-corectComponent[i].JPEGLSPredict(imageData[j-countComponent], 0, 0, 0, imageData[j]);
    j=row_width-countComponent+i;
    dimageData[j]=imageData[j]-corectComponent[i].JPEGLSPredict(imageData[j-countComponent], 0, 0, imageData[j-row_width+countComponent], imageData[j]);
    j=row_width+i;
    dimageData[j]=imageData[j]-corectComponent[i].JPEGLSPredict(imageData[j-countComponent], imageData[j-row_width], 0, imageData[j-row_width+countComponent], imageData[j]);
    for (j=countComponent+row_width+i; j<image_row*row_width; j+=countComponent)
     dimageData[j]=imageData[j]-corectComponent[i].JPEGLSPredict(imageData[j-countComponent], imageData[j-row_width], imageData[j-countComponent-row_width], imageData[j-row_width+countComponent], imageData[j]);
    corectComponent[i].initialize();
   }
  unsigned long freqD[256];
  memset(freqD, 0, sizeof(freqD));
  for (i=0; i<image_row*row_width; i++)
   freqD[dimageData[i]]++;
  double sum=0;
  for (i=0; i<256; i++)
   if (freqD[i]>1)
    sum+=freqD[i]*log(freqD[i]);

  UBYTE1 *dimageData1=new UBYTE1[image_row*row_width];
  for (i=0; i<countComponent;i++)
   {dimageData1[i]=imageData[i];
    for (j=countComponent+i; j<row_width; j+=countComponent)
     dimageData1[j]=imageData[j]-MedPredict(imageData[j-countComponent], 0, 0);
    for (j=row_width+i; j<countComponent+row_width; j+=countComponent)
     dimageData1[j]=imageData[j]-MedPredict(imageData[j-countComponent], imageData[j-row_width], 0);
    for (j=countComponent+row_width+i; j<image_row*row_width; j+=countComponent)
     dimageData1[j]=imageData[j]-MedPredict(imageData[j-countComponent], imageData[j-row_width], imageData[j-countComponent-row_width]);
   }
  memset(freqD, 0, sizeof(freqD));
  for (i=0; i<image_row*row_width; i++)
   freqD[dimageData1[i]]++;
  double sum1=0;
  for (i=0; i<256; i++)
   if (freqD[i]>1)
    sum1+=freqD[i]*log(freqD[i]);
  if (sum1>sum)
   {trebaCorectPredict=false;
    memcpy(dimageData, dimageData1, image_row*row_width);
   }
  delete [] dimageData1;
  dimageData1=NULL;
  //     
  currentPozImage=0;
  countPointImage=image_row*image_col;
  countByteImage=row_width * current_image->Height ();
  globalBlockBuffer=new UBYTE2[2*countPointImage];
  indexPointPaletr=new UBYTE1[countPointImage];
  UBYTE4 *nextPoint=new UBYTE4[countPointImage];
  UBYTE4 p=0;//   
  InitializeHashTable (); //    ProcessImageData
  for (globalBlockBufferCount=0;
       currentPozImage < countByteImage ;
       ++globalBlockBufferCount)
   {unsigned int length, offset, zmLen;
    LongestMatch (length, offset);
    zmLen=0;
    if (compressionVertical && length<PngLongestLength)
     {unsigned int vlength, voffset;
      VLongestMatch (vlength, voffset);
      if (vlength>length || (vlength==length && voffset<offset))
       {if (vlength==PngLongestLength)
         vlength-=3;
        if (vlength>=length)
         {length=vlength;
          offset=voffset;
          zmLen=1;
         }
       }
     }
    if (compressionGorizontal && length<PngLongestLength)
     {unsigned int glength, goffset;
      GLongestMatch (glength, goffset);
      if (glength>length || (glength==length && goffset<offset))
       {if (glength==PngLongestLength)
         glength-=3;
        if (glength>=length)
         {length=glength;
          offset=goffset;
          zmLen=2;
         }
       }
     }
    if (length == 0)
     {unsigned int hashvalue = HashValue (currentPozImage) ;
      FreePozHashEntry(currentPozImage);
      MoveHashEntry (currentPozImage, hashvalue) ;
      if (compressionVertical)
       {hashvalue = VHashValue (currentPozImage) ;
        VFreePozHashEntry(currentPozImage);
        VMoveHashEntry (currentPozImage, hashvalue) ;
       }
      if (compressionGorizontal)
       {hashvalue = GHashValue (currentPozImage) ;
        GFreePozHashEntry(currentPozImage);
        GMoveHashEntry (currentPozImage, hashvalue) ;
       }
      nextPoint[p]=currentPozImage/3;
      p=nextPoint[p];
      globalBlockBuffer[globalBlockBufferCount] = 0;
      currentPozImage+=3;
     }
    else
     {if (trebaOptimizeLZBlock)
       {if (length<minLenZaminaLiteral)
         for (unsigned int ii = 0 ; ii < length/3 ; ii++)
          {nextPoint[p]=currentPozImage/3+ii;
           p=nextPoint[p];
          }
       }
      // We have found a match.  First update the hash table and
      // copy the data in the LZ window.
      for (unsigned int ii = 0 ; ii < length ; ii+=3)
       {FreePozHashEntry(currentPozImage);
        unsigned int hashvalue = HashValue (currentPozImage) ;
        MoveHashEntry (currentPozImage, hashvalue) ;
        if (compressionVertical)
         {hashvalue = VHashValue (currentPozImage) ;
          VFreePozHashEntry(currentPozImage);
          VMoveHashEntry (currentPozImage, hashvalue) ;
         }
        if (compressionGorizontal)
         {hashvalue = GHashValue (currentPozImage) ;
          GFreePozHashEntry(currentPozImage);
          GMoveHashEntry (currentPozImage, hashvalue) ;
         }
        currentPozImage+=3;
       }
      globalBlockBuffer[globalBlockBufferCount] = 256 + length + zmLen;
      ++ globalBlockBufferCount;
      globalBlockBuffer[globalBlockBufferCount] = offset ;
     }
   }
  nextPoint[p]=0;
  generatePaletr(dimageData, nextPoint, indexPointPaletr);
  delete nextPoint;
  nextPoint=NULL;
  initBufPlus();
  currentPointWrite=0;
  output_buffer = new UBYTE1 [OutputBufferSize] ;
  // Fill in the image header.
  PngImageHeader header ;
  header.width = SystemToBigEndian ((UBYTE4) current_image->Width ()) ;
  header.height = SystemToBigEndian ((UBYTE4) current_image->Height ()) ;
  header.bitdepth = current_image->BitCount () ;
  if (current_image->BitCount () != 24)
  {
    // Determine if the image is grayscale.
    bool isgrayscale = true ;
    for (unsigned int ii = 0 ; ii < current_image->ColorCount () ; ++ ii)
    {
      if (current_image->ColorMap (ii).red != current_image->ColorMap (ii).green
          ||
          current_image->ColorMap (ii).red != current_image->ColorMap (ii).blue)
      {
        isgrayscale = false ;
        break ;
      }
    }

    if (isgrayscale)
      header.colortype = Grayscale ;
    else
      header.colortype = Palette ;
  }
  else
  {
    header.bitdepth = 8 ;
    header.colortype = RGB ;
  }
  header.compressionmethod = 1 ; //  
  header.filtermethod = 0 ;
  header.interlacemethod = 0 ;

  // Output the PNG Signature and header.
  output_stream->write ((char *) PngSignature, PngSignatureSize) ;
  StartChunk ("IHDR") ;
  OutputBlock (&header, sizeof (header)) ;
  WriteCurrentChunk () ;
  writeCM();
  writeNotCP();

  // Output any text blocks with keywords defined.
  //WriteTextBlocks () ;

  WriteImageData () ;
  delete globalBlockBuffer;
  globalBlockBuffer=NULL;
  delete indexPointPaletr;
  indexPointPaletr=NULL;
  //    main()
  // Write the IEND marker.
  StartChunk ("IEND") ;
  WriteCurrentChunk () ;

  return ;
}

//
//  Description:
//
//    This function initializes the Hash Table.
//
//
void PngEncoder::InitializeHashTable ()
{
  UBYTE4 ii ;
  memset (hash_values, 0, sizeof (HashEntry) * PngWindowSize) ;
  memset (hash_table, 0, sizeof (PHashEntry) * HashTableSize) ;
  memset (countHashValue, 0, sizeof (unsigned long) * HashTableSize) ;
  if (compressionVertical)
   {memset (vhash_values, 0, sizeof (HashEntry) * PngWindowSize) ;
    memset (vhash_table, 0, sizeof (PHashEntry) * HashTableSize) ;
    memset (vcountHashValue, 0, sizeof (unsigned long) * HashTableSize) ;
   }
  if (compressionGorizontal)
   {memset (ghash_values, 0, sizeof (HashEntry) * PngWindowSize) ;
    memset (ghash_table, 0, sizeof (PHashEntry) * HashTableSize) ;
    memset (gcountHashValue, 0, sizeof (unsigned long) * HashTableSize) ;
   }
  return ;
}

//
//  Description:
//
//    This function writes a Deflate block header and Huffman table
//    descriptions to the output stream.
//
//  Parameters:
//    lastblock: true => This is the last block in the image
//               false => There are more blocks to come
//
void PngEncoder::OutputDeflateHeader (bool lastblock)
{
  // Determine the count of length/literal and distances that
  // are used.
  unsigned int lengthcount ;
  for (lengthcount = PngMaxLengthCodes ;
       lengthcount > 0 ;
       -- lengthcount)
  {
    if (length_table->ehufsi [lengthcount - 1] != 0)
      break ;
  }

  unsigned int distancecount ;
  for (distancecount = PngMaxLengthCodes ;
       distancecount > 1 ;
       -- distancecount)
  {
    if (distance_table->ehufsi [distancecount - 1] != 0)
      break ;
  }

  // Gather the Huffman statistics for encoding the
  // lengths then create the Huffman table for doing so.
  length_length_table->Reset () ;
  FindLengthCodes (&PngEncoder::GatherLengthCounts,
                   length_table->ehufsi,
                   lengthcount) ;
  FindLengthCodes (&PngEncoder::GatherLengthCounts,
                   distance_table->ehufsi,
                   distancecount) ;
  length_length_table->BuildTable (PngMaxLengthLengthCodeSize) ;

  // Count the number of lengths we have to output.
  unsigned int hclen ;
  for (hclen = 19 ; hclen > 0 ; -- hclen)
  {
    if (length_length_table->ehufsi [PngLengthOrder [hclen-1]] != 0)
      break ;
  }

  // Write the Deflate header to the IDAT bock.
  if (lastblock)
    OutputDataBits (1, 1) ;
  else
    OutputDataBits (0, 1) ;

  OutputDataBits (2, 2) ; // Dynamic Huffman Codes
  OutputDataBits (lengthcount - 257, 5) ;
  OutputDataBits (distancecount - 1, 5) ;
  OutputDataBits (hclen - 4, 4) ;
  // Output the data for the Huffman table that encodes the Huffman tables
  for (unsigned int ii = 0 ; ii < hclen ; ++ ii)
    OutputDataBits (length_length_table->ehufsi [PngLengthOrder [ii]], 3) ;

   // Huffman encode the lengths for the Length/Literal and Distance
  // Huffman tables.
  FindLengthCodes (&PngEncoder::OutputLengthCounts,
                   length_table->ehufsi,
                   lengthcount) ;
  FindLengthCodes (&PngEncoder::OutputLengthCounts,
                   distance_table->ehufsi,
                   distancecount) ;
  return ;
}

//
//  Description:
//
//    This function writes a ZLIB header to the output stream.
//
void PngEncoder::OutputZLibHeader ()
{ UBYTE1 stepin=8;
  while ((1<<stepin)<PngWindowSize)
   stepin++;
  UBYTE1 cmf = (stepin-8)*16+8 ;  //   78: 7=>32K Sliding Window, 8=> Deflate Compression
  UBYTE1 flg = compression_level << 6 ;
  UBYTE2 check = (cmf << 8) | flg ;
  flg |= 31 - (check % 31) ;
  OutputDataBits (cmf, 8) ;
  OutputDataBits (flg, 8) ;
  return ;
}

//
//  Description:
//
//    This function writes the image data to a series of IDAT blocks.
//
void PngEncoder::WriteImageData ()
{
  // Write the IDAT blocks.
  StartChunk ("IDAT") ;

  adler_value = 1 ;

  block_buffer = new UBYTE2 [block_buffer_size] ;
  trebaZaminaBuffer = new UBYTE1 [block_buffer_size] ;
  if (trebaOptimizeLZBlock)
   {pozAnalizZamina = new UBYTE4 [block_buffer_size] ;
    pozImageBuffer = new UBYTE4 [block_buffer_size] ;
   }
  else
   memset(trebaZaminaBuffer, 1, block_buffer_size*sizeof(UBYTE1));
  currentPozImage=0;
  currentGlobalPoz=0;
  OutputZLibHeader () ;

  CallProgressFunction (0) ;

  // What we are doing here is outputing the image into a series of
  // Deflate compressed blocks.
  for (bool nomoredata = false ; ! nomoredata ;)
  {
    // Save the input state
    length_table->Reset () ;
    distance_table->Reset () ;
    countAnalizZamina=0;
    nomoredata = ProcessImageData () ;
    if (trebaOptimizeLZBlock)
     {PngHuffmanEncoder statEncoder;
      unsigned int freqCodeLengthTable[PngMaxLengthCodes], freqDistanceLengthTable[PngMaxDistanceCodes];
      memcpy(freqCodeLengthTable, length_table->frequencies, PngMaxLengthCodes*sizeof(unsigned int));
      memcpy(freqDistanceLengthTable, distance_table->frequencies, PngMaxDistanceCodes*sizeof(unsigned int));
      unsigned int len, zmLen, offset, i, j, k, countBitZamina;
      unsigned int code, extra , value, codeD, extraD , valueD;
      UBYTE4 poz;
      UBYTE1 lenTrebaZamina=minLenZaminaLiteral, pprNayavno=0, countIter, maxIter;
      bool nayavnoLenZamina[minLenZaminaLiteral]; //      
      UBYTE4 freqProbaLength[minLenZaminaLiteral][286];
      UBYTE4 freqProbaDistance[minLenZaminaLiteral][30];
      long plusBit[minLenZaminaLiteral+1];
      memset(nayavnoLenZamina, 0, sizeof(nayavnoLenZamina));
      memset(freqProbaLength, 0, sizeof(freqProbaLength));
      memset(freqProbaDistance, 0, sizeof(freqProbaDistance));
      memset(plusBit, 0, sizeof(plusBit));
      //     
      for (i=0; i<countAnalizZamina; i++)
       {len = block_buffer[pozAnalizZamina[i]] - 256;
        zmLen=len%3;
        len-=zmLen;
        offset=block_buffer[pozAnalizZamina[i]+1];
        LengthToCode (len+zmLen, code, extra, value);
        DistanceToCode (offset, codeD, extraD, valueD) ;
        poz=pozImageBuffer[pozAnalizZamina[i]];
        nayavnoLenZamina[len]=true;
        //     
        for (j=0; j<len/3; j++)
         {freqProbaLength[len][indexPointPaletr[poz+j]]++;
          plusBit[len]-=(countBitRColorPaletr[indexPointPaletr[poz+j]]+
                         countBitGColorPaletr[indexPointPaletr[poz+j]]+
                         countBitBColorPaletr[indexPointPaletr[poz+j]]);
         }
        freqProbaLength[len][code]++;
        freqProbaDistance[len][codeD]++;
        plusBit[len]+=extra+extraD;
       }
      double minSizeCode=0xFFFFFFFF, sizeCode, sizeLength, sizeDistance;
      UBYTE4 newCountAnalizZamina, countLength, countDistance, sizeLiteral, sizeZamina, freq;
      nayavnoLenZamina[2]=true; //      
      UBYTE1 lenProbaLength[286];
      UBYTE1 lenProbaDistance[30];
      unsigned int freqEncoder[286];
      for (k=2; k<minLenZaminaLiteral; k++) //      
       if (nayavnoLenZamina[k])
        {countLength=0;
         sizeLength=0;
         for (i=0; i<=256; i++)
          {freqProbaLength[k][i]+=freqProbaLength[pprNayavno][i]; //  ,  
           freq=freqCodeLengthTable[i]+freqProbaLength[k][i];
           countLength+=freq;
           if (freq>1)
            sizeLength+=freq*log(freq);
          }
         for (i=257; i<=285; i++)
          {freqProbaLength[k][i]+=freqProbaLength[pprNayavno][i]; //  ,  
           freq=freqCodeLengthTable[i]-freqProbaLength[k][i];
           countLength+=freq;
           if (freq>1)
            sizeLength+=freq*log(freq);
          }
         sizeCode=countLength*log(countLength)-sizeLength;
         countDistance=0;
         sizeDistance=0;
         for (i=0; i<=29; i++)
          {freqProbaDistance[k][i]+=freqProbaDistance[pprNayavno][i]; //  ,  
           freq=freqDistanceLengthTable[i]-freqProbaDistance[k][i];
           countDistance+=freq;
           if (freq>1)
            sizeDistance+=freq*log(freq);
          }
         if (countDistance)
          sizeCode+=countDistance*log(countDistance)-sizeDistance;
         plusBit[k]+=plusBit[pprNayavno]; //  ,  
         sizeCode=sizeCode/log(2)-plusBit[k];
         if (sizeCode<minSizeCode)
          {minSizeCode=sizeCode;
           lenTrebaZamina=k;
          }
         pprNayavno=k;
        }
      //    
      UBYTE1 maxLenLength, maxLenDistance;
      statEncoder.Reset();
      for (i=0; i<=256; i++)
       statEncoder.frequencies[i]=freqCodeLengthTable[i]+freqProbaLength[lenTrebaZamina][i];
      for (i=257; i<=285; i++)
       statEncoder.frequencies[i]=freqCodeLengthTable[i]-freqProbaLength[lenTrebaZamina][i];
      statEncoder.BuildTable(PngMaxLengthCodeSize);
      maxLenLength=statEncoder.ehufsi[256];
      for (i=0; i<=285; i++)
       if (statEncoder.ehufsi[i])
        lenProbaLength[i]=statEncoder.ehufsi[i];
       else
        lenProbaLength[i]=maxLenLength;
      statEncoder.Reset();
      for (i=0; i<=29; i++)
       statEncoder.frequencies[i]=freqDistanceLengthTable[i]-freqProbaDistance[lenTrebaZamina][i];
      statEncoder.frequencies[30]=1; //    
      statEncoder.BuildTable(PngMaxLengthCodeSize);
      maxLenDistance=statEncoder.ehufsi[30];
      if (maxLenDistance<3)
       maxLenDistance=3;
      for (i=0; i<=29; i++)
       if (statEncoder.ehufsi[i])
        lenProbaDistance[i]=statEncoder.ehufsi[i];
       else
        lenProbaDistance[i]=maxLenDistance;
      //     
      for (i=0; i<countAnalizZamina; i++)
       {len = block_buffer[pozAnalizZamina[i]] - 256;
        zmLen=len%3;
        len-=zmLen;
        offset=block_buffer[pozAnalizZamina[i]+1];
        LengthToCode (len+zmLen, code, extra, value);
        DistanceToCode (offset, codeD, extraD, valueD) ;
        poz=pozImageBuffer[pozAnalizZamina[i]];
        sizeZamina=lenProbaLength[code]+extra+lenProbaDistance[codeD]+extraD;
        sizeLiteral=0;
        for (j=0; j<len/3 && sizeLiteral<sizeZamina; j++)
         sizeLiteral+=lenProbaLength[indexPointPaletr[poz+j]]+
                      (countBitRColorPaletr[indexPointPaletr[poz+j]]+
                       countBitGColorPaletr[indexPointPaletr[poz+j]]+
                       countBitBColorPaletr[indexPointPaletr[poz+j]]);
        if (sizeLiteral<sizeZamina)
         {// 
          trebaZaminaBuffer[pozAnalizZamina[i]]=0;
          for (j=0; j<len/3; j++)
           freqCodeLengthTable[indexPointPaletr[poz+j]]++;
          freqCodeLengthTable[code]--;
          freqDistanceLengthTable[codeD]--;
          switch (zmLen)
           {case 0:countZamina--;
                   break;
            case 1:countZaminaV--;
                   break;
            case 2:countZaminaG--;
           }
         }
       }
      //  
      maxIter=1;
      if (lenTrebaZamina==2)
       maxIter=0;
      countIter=0;
      while (countIter<maxIter)
       {countIter++;
        //    
        statEncoder.Reset();
        for (i=0; i<=285; i++)
         statEncoder.frequencies[i]=freqCodeLengthTable[i];
        statEncoder.BuildTable(PngMaxLengthCodeSize);
        maxLenLength=statEncoder.ehufsi[256];
        for (i=0; i<=285; i++)
         if (statEncoder.ehufsi[i])
          lenProbaLength[i]=statEncoder.ehufsi[i];
         else
          lenProbaLength[i]=maxLenLength;
        statEncoder.Reset();
        for (i=0; i<=29; i++)
         statEncoder.frequencies[i]=freqDistanceLengthTable[i];
        statEncoder.frequencies[30]=1; //    
        statEncoder.BuildTable(PngMaxLengthCodeSize);
        maxLenDistance=statEncoder.ehufsi[30];
        if (maxLenDistance<3)
         maxLenDistance=3;
        for (i=0; i<=29; i++)
         if (statEncoder.ehufsi[i])
          lenProbaDistance[i]=statEncoder.ehufsi[i];
         else
          lenProbaDistance[i]=maxLenDistance;
        //     
        for (i=0; i<countAnalizZamina; i++)
         {len = block_buffer[pozAnalizZamina[i]] - 256;
          zmLen=len%3;
          len-=zmLen;
          offset=block_buffer[pozAnalizZamina[i]+1];
          LengthToCode (len+zmLen, code, extra, value);
          DistanceToCode (offset, codeD, extraD, valueD) ;
          poz=pozImageBuffer[pozAnalizZamina[i]];
          sizeZamina=lenProbaLength[code]+extra+lenProbaDistance[codeD]+extraD;
          sizeLiteral=0;
          if (trebaZaminaBuffer[pozAnalizZamina[i]]) // 
           {for (j=0; j<len/3 && sizeLiteral<sizeZamina; j++)
             sizeLiteral+=lenProbaLength[indexPointPaletr[poz+j]]+
                          (countBitRColorPaletr[indexPointPaletr[poz+j]]+
                           countBitGColorPaletr[indexPointPaletr[poz+j]]+
                           countBitBColorPaletr[indexPointPaletr[poz+j]]);
            if (sizeLiteral<sizeZamina)
             {// 
              trebaZaminaBuffer[pozAnalizZamina[i]]=0;
              for (j=0; j<len/3; j++)
               freqCodeLengthTable[indexPointPaletr[poz+j]]++;
              freqCodeLengthTable[code]--;
              freqDistanceLengthTable[codeD]--;
              switch (zmLen)
               {case 0:countZamina--;
                       break;
                case 1:countZaminaV--;
                       break;
                case 2:countZaminaG--;
               }
             }
           }
          else // 
           {for (j=0; j<len/3; j++)
             sizeLiteral+=lenProbaLength[indexPointPaletr[poz+j]]+
                          (countBitRColorPaletr[indexPointPaletr[poz+j]]+
                           countBitGColorPaletr[indexPointPaletr[poz+j]]+
                           countBitBColorPaletr[indexPointPaletr[poz+j]]);
            if (sizeLiteral>sizeZamina)
             {// 
              trebaZaminaBuffer[pozAnalizZamina[i]]=1;
              for (j=0; j<len/3; j++)
               freqCodeLengthTable[indexPointPaletr[poz+j]]--;
              freqCodeLengthTable[code]++;
              freqDistanceLengthTable[codeD]++;
              switch (zmLen)
               {case 0:countZamina++;
                       break;
                case 1:countZaminaV++;
                       break;
                case 2:countZaminaG++;
               }
             }
           }
         }
       }
      memcpy(length_table->frequencies, freqCodeLengthTable,  PngMaxLengthCodes*sizeof(unsigned int));
      memcpy(distance_table->frequencies, freqDistanceLengthTable, PngMaxDistanceCodes*sizeof(unsigned int));
     }
    // Create the Huffman tables for length/literals and distances.
    length_table->BuildTable (PngMaxLengthCodeSize) ;
    distance_table->BuildTable (PngMaxDistanceCodeSize) ;
    OutputDeflateHeader (nomoredata) ;
    OutputBlockData () ;
    CallProgressFunction (currentPozImage*100/countByteImage);
  }

  CallProgressFunction (100) ;
  OutputAdler () ;
  WriteCurrentChunk () ;

  delete [] block_buffer ; block_buffer = NULL ;
  delete [] trebaZaminaBuffer ; trebaZaminaBuffer = NULL ;
  if (trebaOptimizeLZBlock)
   {delete [] pozAnalizZamina; pozAnalizZamina = NULL;
    delete [] pozImageBuffer; pozImageBuffer = NULL;
   }
  return ;
}

//
//  Description:
//
//    This function Huffman encodes and outputs the buffered data.
//
//    The buffer is encoded so that
//
//      0..255  is a literal byte
//      256-514 is a length code of N-256
//
//    Each length code is followed by a distance code.
//
void PngEncoder::OutputBlockData ()
{
  UBYTE2 huffmancode ;
  UBYTE1 huffmansize ;
  unsigned int code ;
  unsigned int extra ;
  unsigned int value ;
  unsigned int limit = block_buffer_count ;
  for (unsigned int ii = 0 ; ii < limit ; ++ ii)
  { if (block_buffer [ii] < 256)
    { length_table->Encode (block_buffer [ii], huffmancode, huffmansize) ;
      OutputDataBits (huffmancode, huffmansize) ;
      writeBufPlus(dimageData[3*currentPointWrite]-minRColorPaletr[block_buffer[ii]], countBitRColorPaletr[block_buffer[ii]]);
      writeBufPlus(dimageData[3*currentPointWrite+1]-minGColorPaletr[block_buffer[ii]], countBitGColorPaletr[block_buffer[ii]]);
      writeBufPlus(dimageData[3*currentPointWrite+2]-minBColorPaletr[block_buffer[ii]], countBitBColorPaletr[block_buffer[ii]]);
      currentPointWrite++;
    }
    else
    {
      unsigned int length = block_buffer [ii] - 256 ;
      if (trebaZaminaBuffer[ii])
       {LengthToCode (length, code, extra, value) ;
        length_table->Encode (code, huffmancode, huffmansize) ;
        OutputDataBits (huffmancode, huffmansize) ;
        if (extra != 0)
          OutputDataBits (value, extra) ;

        ++ ii ;
        unsigned int distance = block_buffer [ii] ;
        DistanceToCode (distance, code, extra, value) ;
        distance_table->Encode (code, huffmancode, huffmansize) ;
        OutputDataBits (huffmancode, huffmansize) ;
        if (extra != 0)
          OutputDataBits (value, extra) ;
        currentPointWrite+=length/3;
       }
      else
       {UBYTE4 poz=pozImageBuffer[ii];
        length-=length%3;
        for (unsigned int j = 0; j<length/3; j++)
         {length_table->Encode (indexPointPaletr[poz+j], huffmancode, huffmansize) ;
          OutputDataBits (huffmancode, huffmansize) ;
          writeBufPlus(dimageData[3*currentPointWrite]-minRColorPaletr[indexPointPaletr[poz+j]], countBitRColorPaletr[indexPointPaletr[poz+j]]);
          writeBufPlus(dimageData[3*currentPointWrite+1]-minGColorPaletr[indexPointPaletr[poz+j]], countBitGColorPaletr[indexPointPaletr[poz+j]]);
          writeBufPlus(dimageData[3*currentPointWrite+2]-minBColorPaletr[indexPointPaletr[poz+j]], countBitBColorPaletr[indexPointPaletr[poz+j]]);
          currentPointWrite++;
         }
        ++ii;
       }
    }
  }
  length_table->Encode (PngEndCode, huffmancode, huffmansize) ;
  OutputDataBits (huffmancode, huffmansize) ;
  return ;
}










